package de.mospace.apps.metronome;

/* Metronome
 * Copyright (C) 2007 Moritz Ringler
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

import javax.sound.midi.Instrument;
import javax.sound.midi.InvalidMidiDataException;
import javax.sound.midi.MidiEvent;
import javax.sound.midi.MidiSystem;
import javax.sound.midi.MidiUnavailableException;
import javax.sound.midi.Patch;
import javax.sound.midi.Sequence;
import javax.sound.midi.Sequencer;
import javax.sound.midi.ShortMessage;
import javax.sound.midi.Soundbank;
import javax.sound.midi.Synthesizer;
import javax.sound.midi.Track;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;

public class MidiMetronome{
    private static final int CHANNEL = 0;
    private static final int NOTE = 69; //A (440 Hz)
    private static final int VELOCITY = 90;
    private static final int LENGTH_ON = 2;
    private static final int LENGTH_OFF = 15;

    private Sequencer sequencer = MidiSystem.getSequencer();
    private final Synthesizer synth = MidiSystem.getSynthesizer();
    private Soundbank soundbank = synth.getDefaultSoundbank();
    //default to wood block
    private Instrument instrument = getSoundbank().getInstrument(
            new Patch(0, 115));
    private Sequence seq;

    public MidiMetronome() throws MidiUnavailableException{

    }

    private Soundbank getSoundbank(){
        synchronized(this){
            if(soundbank == null){
                try{
                    soundbank = MidiSystem.getSoundbank(
                            MidiMetronome.class.getResource("/soundbank-min.gm"));
                } catch (Exception ex){
                    ex.printStackTrace();
                    System.exit(1);
                }
            }
        }
        return soundbank;
    }

    public void setInstrument(Instrument instr) throws MidiUnavailableException, InvalidMidiDataException{
        if(instrument != instr && instr != null){
            int bpm = isRunning()? ((int) sequencer.getTempoInBPM()) : 0;
            stop();
            if(instrument != null){
                //synth.unloadInstrument(instrument);
            }
            System.gc();
            instrument = instr;
            if(bpm != 0){
                start(bpm);
            }
        }
    }

    public Instrument[] getAvailableInstruments() throws MidiUnavailableException{
        return getSoundbank().getInstruments();
    }

    public Instrument getInstrument(){
        return instrument;
    }

    public void setTempo(int bpm){
        sequencer.setTempoInBPM(bpm);
    }

    public boolean isRunning(){
        return sequencer != null && sequencer.isRunning();
    }


    private Sequence generateTokSequence()
            throws MidiUnavailableException, InvalidMidiDataException{

        int length = LENGTH_ON + LENGTH_OFF;
        Sequence seq = new Sequence(Sequence.PPQ, length);
        Track track = seq.createTrack();
        ShortMessage msg;

        //generate 10000 beeps in a single track
        length = length * 10000;
        track.add(makeInstrumentChange());

        for (int i = 1; i < length; i+= LENGTH_OFF) { // Anzahl der Noten
            // generate NOTE ON message
            msg = new ShortMessage();
            msg.setMessage(ShortMessage.NOTE_ON, CHANNEL, NOTE,  VELOCITY);
            track.add(new MidiEvent(msg, i));

            //generate NOTE_OFF message
            i += LENGTH_ON;
            msg = new ShortMessage();
            msg.setMessage(ShortMessage.NOTE_OFF, CHANNEL, NOTE, 0);
            track.add(new MidiEvent(msg, i));
        }
        return seq;
    }

    public void start(int bpm)  throws MidiUnavailableException, InvalidMidiDataException{

        //open synthesizer
        synth.open();

        //open and connect sequencer
        sequencer = MidiSystem.getSequencer();
        sequencer.open();
        sequencer.getTransmitter().setReceiver(synth.getReceiver());

        //set sequence/track, generate if necessary
        synchronized(sequencer){
            if(seq == null){
                seq = generateTokSequence();
            }
        }
        sequencer.setSequence(seq);
        sequencer.setLoopCount(Sequencer.LOOP_CONTINUOUSLY);

        //update the midi event that selects the instrument
        //it is the first event of the track at tick 0
        Track track = sequencer.getSequence().getTracks()[0];
        track.remove(track.get(0));
        track.add(makeInstrumentChange());

        //set tempo
        sequencer.setTempoInBPM(bpm);

        //start playing
        sequencer.start();
    }

    private MidiEvent makeInstrumentChange()   throws MidiUnavailableException, InvalidMidiDataException{
        ShortMessage msg = new ShortMessage();
        System.err.println("Loading instrument " + instrument);
        //Kanal CHANNEL auf aktuelles Instrument umschalten
        Patch p = instrument.getPatch();
        msg.setMessage(ShortMessage.PROGRAM_CHANGE, CHANNEL,
                        p.getProgram(), p.getBank());
        return new MidiEvent(msg, 0);
    }

    public void stop(){
        if(isRunning()){
            sequencer.stop();
            sequencer.close();
            synth.close();
        }
    }

    public static void main(String[] argv) throws Exception{
        if(argv.length > 0){
            //run in command line mode using first argument as bpm
            //and default instrument
            (new MidiMetronome()).start(Integer.parseInt(argv[0]));
        } else {
            //run GUI
            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
            SwingUtilities.invokeLater(new Runnable()
            {
                public void run()
                {
                    try{
                        MetronomeGUI gui = new MetronomeGUI();
                        gui.pack();
                        gui.setVisible(true);
                    } catch (Exception ex){
                        System.err.println(ex);
                        System.exit(1);
                    }
                }
            });
        }
    }

}
